; ==================================================================================================
; Title:   XPropertyTree.inc
; Author:  G. Friedrich
; Version: 1.0.0
; Purpose: Customized XTreeView and XTreeViewNode to be used to show and edit properties.
; Notes:   Version 1.0.0, August 2007
;            - First release.
; ==================================================================================================

; Features:
; 
; Same as XTreeView.
; Support for strings/32 bit integers/real4 floating point numbers/paths/files/colors.


; Some rules:
; 
; 1. See XTreeView object.
; 2. Each node (Category or Property) need to be created, initialized and finally inserted.
;      New XPT_PropertyNode, Init, pOwner, hIcon, "Description", dHeight, XTVNS_EXPANDED, \
;          pVariable, dNodeType, dOptions
;      OCall ebx::XPropertyTree.Insert, eax, pCategoryNode, XTVN_INSERT_AS_LAST_CHILD
; 3. Property nodes of type XPT_VAR_TYPE_STRING passes the max string size including the
;    terminating zero.
; 4. Property nodes of type XPT_VAR_TYPE_COLOR passes a CHOOSECOLOR structure. Following
;    members must be initialized before call: hInstance, Flags, lCustData, lpfnHook,
;   lpTemplateName.
; 5. Property nodes of type XPT_VAR_TYPE_FILE passes a OPENFILENAME structure. Following
;    members must be initialized before call: hInstance, lpstrFilter, lpstrCustomFilter,
;    nMaxCustFilter, nFilterIndex, lpstrFile, nMaxFile, lpstrFileTitle, nMaxFileTitle,
;    lpstrInitialDir, lpstrTitle, Flags, lpstrDefExt.
;    ComDlg32 library must be loaded.
; 6. Property nodes of type XPT_VAR_TYPE_PATH passes a BROWSEINFO structure. Following
;    members must be initialized before call: pidlRoot, lpszTitle, ulFlags.
;    Ole32 library must be loaded.
; 7. Property nodes of type XPT_VAR_TYPE_LIST passes a doble zero terminating string
;    options list.
; 8. All other property nodes don't pass any argument.


XPT_ICON_SIZE                       equ   16
XPT_COLOR_RECT_SIZE                 equ   12
XPT_MARGIN                          equ   03
XPT_FONT_SIZE_CAT                   equ   10
XPT_FONT_SIZE_PROP                  equ   10
XPT_CONTROL_ID_EDIT                 equ   120
XPT_CONTROL_ID_BUTTON               equ   121
XPT_CONTROL_ID_COMBOBOX             equ   122

XPT_FLAGS                           equ   XTVF_SHOW_TOGGLER
XPT_PROPERTY_NAME_WIDTH             equ   100

XPT_VAR_TYPE_STRING                 equ   0         ;pOptions -> Max string size inc. term. zero
XPT_VAR_TYPE_INTEGER                equ   1         ;pOptions -> NULL
XPT_VAR_TYPE_REAL4                  equ   2         ;pOptions -> NULL
XPT_VAR_TYPE_COLOR                  equ   3         ;pOptions -> CHOOSECOLOR
XPT_VAR_TYPE_SYS_COLOR              equ   4         ;pOptions -> NULL
XPT_VAR_TYPE_FILE                   equ   5         ;pOptions -> OPENFILENAME
XPT_VAR_TYPE_PATH                   equ   6         ;pOptions -> BROWSEINFO
XPT_VAR_TYPE_LIST                   equ   7         ;pOptions -> Double zero terminated list


XPTS_CHILD_FOCUSED                  equ   BIT07



;  Node layout
;  
;                                            ContentRect(*)
;   
;  |                         |
;  |                       |  ||
;  |                  [-]---|| Icon | Caption | Text(*)   | Button |||
;  |                Toggler |  ||
;  |                 Rect     |
;   
;
; Note: (*) structures contained in XTVN_DISP_INFO
;


XPT_DISP_INFO struct            ;Extend the XTVN_DISP_INFO structure
              XTVN_DISP_INFO    {}      ;Anonymous definition
  IconRect    RECT              {}      ;RECT structure for the node icon
  CaptionRect RECT              {}      ;RECT structure for the node caption
  ButtonRect  RECT              {}      ;RECT struc for the control on the right side of the node
XPT_DISP_INFO ends


Object XPropertyTree, 0, XTreeView
  RedefineMethod    Init,               Pointer, Handle, Pointer
  RedefineMethod    Draw
  StaticMethod      OnCommand,          dword, dword
  RedefineMethod    OnCreate,           dword, dword
  RedefineMethod    OnDestroy,          dword, dword
  StaticMethod      OnDrawItem,         dword, dword
  RedefineMethod    OnKeyDown,          dword, dword
  StaticMethod      OnMeasureItem,      dword, dword
  RedefineMethod    OnMouseWheel,       dword, dword
  RedefineMethod    OnSize,             dword, dword
  RedefineMethod    OnVScroll,          dword, dword
  RedefineMethod    RenderBackground,   Pointer                     ;-> RECT
  RedefineMethod    Startup                                         ;Register new wnd class
  RedefineMethod    WndProc,            dword, dword, dword         ;Create new wnd class

  Event     OnCommand,          WM_COMMAND
  Event     OnDrawItem,         WM_DRAWITEM
  Event     OnMeasureItem,      WM_MEASUREITEM

  RedefineVariable  dDispInfoSize,      sizeof XPT_DISP_INFO        ;Extended structure
  RedefineVariable  dFlags,             XPT_FLAGS
  RedefineVariable  dNodeIndent,        12
  RedefineVariable  dColorBackground,   $RGB(224,224,224)
  DefineVariable    dColorBackgroundDark,   dword,    $RGB(192,192,192)

  DefineVariable    pEdtCtrl,           Pointer,      NULL          ;-> EDIT control object
  DefineVariable    pBtnCtrl,           Pointer,      NULL          ;-> BUTTON control object
  DefineVariable    pCBxCtrl,           Pointer,      NULL          ;-> COMBOBOX control object
ObjectEnd


Object XPT_CategoryNode, 0, XTreeViewNode
  RedefineMethod    Done
  RedefineMethod    Init,               Pointer, Handle, Pointer, dword, dword
  RedefineMethod    CanDragMe,          dword
  RedefineMethod    MouseHitDbl,        dword, Pointer, Pointer, dword
  RedefineMethod    RenderContent,      Pointer                     ;Render check box

  RedefineVariable  dState,             XTVS_VISIBLE or XTVS_DISABLED
  DefineVariable    hIcon,              Handle,       0
  DefineVariable    pCaption,           Pointer,      NULL
  DefineVariable    hFont,              Handle,       0
ObjectEnd


Object XPT_PropertyNode, 0, XTreeViewNode
  RedefineMethod    Done
  RedefineMethod    Init,               Pointer, Handle, Pointer, dword, dword, Pointer, dword, Pointer
  RedefineMethod    CanDragMe,          dword
  RedefineMethod    MouseHit,           dword, Pointer, Pointer, dword
  RedefineMethod    MouseHitDbl,        dword, Pointer, Pointer, dword
  RedefineMethod    Render,             Pointer                     ;-> XTVN_DISP_INFO
  RedefineMethod    RenderContent,      Pointer                     ;Render check box

  DefineVariable    hIcon,              Handle,       0
  DefineVariable    hFont,              Handle,       0
  DefineVariable    dFontHeight,        dWord,        0
  DefineVariable    dFontAveWidth,      dWord,        0
  DefineVariable    pCaption,           Pointer,      NULL
  DefineVariable    pVariable,          Pointer,      NULL
  DefineVariable    dVarType,           dword,        0
  DefineVariable    pCtrlType,          Pointer,      NULL
  DefineVariable    pCtrlOptions,       Pointer,      NULL
ObjectEnd


XPT_ControlDef struc
  dCtlID      dword     ?   ;Contol ID
  dStyle      dword     ?   ;Style
  dExStyle    dword     ?   ;Extended style
  pTitle      Pointer   ?   ;Window title
  sdPosX      sdword    ?   ;X position
  sdPosY      sdword    ?   ;Y position
  dWidth      dword     ?   ;Width
  dHeight     dword     ?   ;Height
XPT_ControlDef ends

Object XPT_Edit, 0, WinControl
  RedefineMethod    Done
  RedefineMethod    Init,               Pointer, Handle, Pointer
  StaticMethod      OnKeyDown,          dword, dword
  RedefineMethod    WndProc,            dword, dword, dword

  Event     OnKeyDown,          WM_KEYDOWN
ObjectEnd

Object XPT_Button, 0, WinControl
  RedefineMethod    Done
  RedefineMethod    Init,               Pointer, Handle, Pointer
  StaticMethod      OnKeyDown,          dword, dword
  RedefineMethod    WndProc,            dword, dword, dword

  Event     OnKeyDown,          WM_KEYDOWN
ObjectEnd

XPTCB_ESCAPING    equ   BIT00

Object XPT_ComboBox, 0, WinControl
  RedefineMethod    Done
  RedefineMethod    Init,               Pointer, Handle, Pointer
  StaticMethod      OnKeyDown,          dword, dword
  StaticMethod      OnPaint,            dword, dword
  RedefineMethod    WndProc,            dword, dword, dword

  Event     OnKeyDown,          WM_KEYDOWN
  Event     OnPaint,            WM_PAINT

  DefineVariable    dState,             dword,        0
ObjectEnd

; ==================================================================================================

if IMPLEMENT

.data
XPT_LogFontText LOGFONT <0, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, DEFAULT_CHARSET, \
                         OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, CLEARTYPE_QUALITY, \
                         DEFAULT_PITCH or FF_DONTCARE, "Arial">

CStr szXPT_Class, "XPropertyTree"

.const

XPT_SYS_COLOR_TABLE_SIZE  equ 30

XPT_SYS_COLOR_STR label byte
byte "Scrollbar", 0
byte "Background / Desktop", 0
byte "Active caption", 0
byte "Inactive caption", 0
byte "Menu", 0
byte "Window", 0
byte "Window frame", 0
byte "Menu text", 0
byte "Window text", 0
byte "Caption text", 0
byte "Active border", 0
byte "Inactive border", 0
byte "App workspace", 0
byte "Highlight", 0
byte "Highlight text", 0
byte "Button face / 3D face", 0
byte "Button shadow / 3D shadow", 0
byte "Gray text", 0
byte "Button text", 0
byte "Inactive caption text", 0
byte "3D highlight / Button highlight", 0
byte "3D dark shadow", 0
byte "3D light", 0
byte "Info text", 0
byte "Info background", 0
byte "Hot light", 0
byte "Gadient active caption", 0
byte "Gadient inactive caption", 0
byte "Menu highlight", 0
byte "Menubar", 0, 0                        ;Double zero signaling the end of the list

XPT_SYS_COLOR_STR_ERROR label byte
byte "Invalid selection", 0

XPT_SYS_COLOR_VAL label byte
byte 00, 01, 02, 03, 04, 05, 06, 07, 08, 09, 10, 11, 12, 13, 14
byte 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 26, 27, 28, 29, 30

.code

; ==================================================================================================
;     XPT_CategoryNode
; ==================================================================================================


;##\

; 
; Method:    XPT_CategoryNode.Done
; Purpose:   Finalizes the XPT_CategoryNode object and all childs in its branch.
; Arguments: None.
; Return:    Nothing.

Method XPT_CategoryNode.Done, uses esi
    ;DbgText "XPT_CategoryNode.Done"
    SetObject esi
    invoke DeleteObject, [esi].hFont
    ACall esi.Done
MethodEnd, NONE

; 
; Method:    XPT_CategoryNode.Init
; Purpose:   Initializes a XPT_CategoryNode object.
; Arguments: Arg1: -> Owner object, XTreeView.
;            Arg2: Icon Handle.
;            Arg2: -> Node Text. Must remain valid for the lifetime of the object instance.
;            Arg3: Node height.
;            Arg4: Node state (XTVNS_DISABLED, XTVNS_EXPANDED, etc.
; Return:    Nothing.

Method XPT_CategoryNode.Init, uses edi esi, pOwner:Pointer, hIcon:Handle, pCaption:Pointer, \
                                     dHeight:dword, dState:dword
    local hDC:Handle

    SetObject esi
    ACall esi.Init, pOwner, 0, dHeight, dState
    m2m [esi].hIcon, hIcon, eax
    m2m [esi].pCaption, pCaption, eax

    ;Create the text font
    mov edi, [esi].pOwner
    mov hDC, $invoke(GetDC, [edi].XPropertyTree.hWnd)
    invoke GetDeviceCaps, hDC, LOGPIXELSY
    invoke MulDiv, XPT_FONT_SIZE_CAT, eax, -72
    mov XPT_LogFontText.lfHeight, eax
    mov XPT_LogFontText.lfWeight, FW_BOLD
    mov [esi].hFont, $invoke(CreateFontIndirect, offset XPT_LogFontText)
    invoke ReleaseDC, [edi].XPropertyTree.hWnd, hDC

MethodEnd, NONE

; 
; Method:    XPT_CategoryNode.CanDragMe
; Purpose:   Indicates if a drag operation can be started.
; Arguments: Arg1: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    Always FALSE => no dragging allowed.

Method XPT_CategoryNode.CanDragMe, NOFRAME, dKeyFlags:dword
    xor eax, eax
MethodEnd 8, eax

; 
; Method:    XPT_CategoryNode.MouseHitDbl
; Purpose:   Checks if the mouse has double clicked a hotspot and performs an action. In this case,
;            the node is toggled.
; Arguments: Arg1: button that triggered this action.
;            Arg2: -> clicked point.
;            Arg3: -> XTVN_DISP_INFO.
;            Arg4: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    eax = TRUE if and action was triggered, otherwise FALSE.

Method XPT_CategoryNode.MouseHitDbl, uses ebx esi, dButton:dword, pPoint:Pointer, \
                                     pDispInfo:Pointer, dKeyFlags:dword
    SetObject esi
    OCall esi.Toggle
    mov ecx, [esi].pOwner
    mov ebx, [ecx].XPropertyTree.hWnd
    OCall ecx::XPropertyTree.Draw
    invoke InvalidateRect, ebx, NULL, FALSE
    invoke UpdateWindow, ebx
    xor eax, eax                                 
MethodEnd, eax

; 
; Method:    XPT_CategoryNode.RenderContent
; Purpose:   Draws the content (icon + caption) of the node.
; Arguments: Arg1: -> XPT_DISP_INFO
; Return:    Nothing.
; Note:      This method must update the coords of all Rect structures in the Content and must
;            invalidate the modified region.

Method XPT_CategoryNode.RenderContent, uses ebx edi esi, pDispInfo:Pointer
    local dPrevColor:dword, dPrevBkColor:dword, dPrevBkMode:dword, InvRect:RECT, hFontPrev:Handle

    ;DbgText "XPT_CategoryNode.RenderContent"
    mov edi, pDispInfo
    .if edi != NULL
      SetObject esi
      mov ebx, [esi].pOwner
      assume edi:ptr XPT_DISP_INFO, ebx:ptr XPropertyTree
      m2m [edi].TextRect.left, [edi].ContentRect.left, eax
      m2m [edi].TextRect.right, [edi].ContentRect.right, eax

      ;Get the drawing rect
      m2m InvRect.left, [edi].ContentRect.left, eax
      m2m InvRect.top, [edi].sdNodeTop, eax
      m2m InvRect.right, [ebx].ClientSize.x, eax
      m2m InvRect.bottom, [edi].sdNodeBot, eax

      ;Render the background
      OCall ebx::XPropertyTree.RenderBackground, addr InvRect

      ;Calc icon position
      mov eax, [edi].ContentRect.bottom
      sub eax, [edi].ContentRect.top
      sub eax, XPT_ICON_SIZE
      shr eax, 1
      mov ecx, [edi].ContentRect.bottom
      sub ecx, eax
      sub ecx, XPT_ICON_SIZE
      mov [edi].IconRect.top, ecx                     ;ecx = Icon top

      mov edx, [edi].ContentRect.left
      add edx, 3                                      ;edx = Icon left
      mov eax, edx
      add eax, 3
      add eax, XPT_ICON_SIZE
      mov [edi].TextRect.left, eax                    ;TextRect.left

      ;Render the icon
      invoke DrawIconEx, [ebx].hMemDC, edx, ecx, [esi].hIcon, \
                         XPT_ICON_SIZE, XPT_ICON_SIZE, NULL, NULL, DI_NORMAL

      ;Set text and background colors
      ;  Check if the whole XTV or the single node is disabled
      .if ([ebx].dState & XTVS_DISABLED) || ([esi].dState & XTVNS_DISABLED)
        ;Disabled
        invoke GetSysColor, COLOR_BTNHILIGHT
        mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
        mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
      .else
        ;Enabled
        .if esi == [ebx].pSelectedNode
          .if ([ebx].dState & XTVS_FOCUSED) || ([ebx].dState & XPTS_CHILD_FOCUSED)
            .ifBitSet [esi].dState, XTVNS_HOTTRACK      ;Check if the node is beeing hottracked
              mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, $RGB(255,255,0))
            .else
              invoke GetSysColor, COLOR_HIGHLIGHTTEXT
              mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
            .endif
            invoke GetSysColor, COLOR_HIGHLIGHT
            mov dPrevBkColor, $invoke(SetBkColor, [ebx].hMemDC, eax)
            mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, OPAQUE)
          .else
            invoke GetSysColor, COLOR_WINDOWTEXT
            mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
            invoke GetSysColor, COLOR_BTNFACE
            mov dPrevBkColor, $invoke(SetBkColor, [ebx].hMemDC, eax)
            mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, OPAQUE)
          .endif
        .elseIfBitSet [esi].dState, XTVNS_HOTTRACK      ;Check if the node is beeing hottracked
          mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, $RGB(0,0,255))
          mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
        .else
          invoke GetSysColor, COLOR_WINDOWTEXT
          mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
          mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
        .endif
      .endif

      ;Set the font
      mov hFontPrev, $invoke(SelectObject, [ebx].hMemDC, [esi].hFont)

      ;Render the created text
      OCall esi.RenderText, edi, [esi].pCaption

      ;Restore previous settings
      invoke SelectObject, [ebx].hMemDC, hFontPrev
      invoke SetBkMode, [ebx].hMemDC, dPrevBkMode
      invoke SetBkColor, [ebx].hMemDC, dPrevBkColor
      invoke SetTextColor, [ebx].hMemDC, dPrevColor

      ;Invalidate the new drawn rect
      invoke InvalidateRect, [ebx].hWnd, addr InvRect, FALSE
      assume edi:NOTHING, ebx:NOTHING
    .endif
MethodEnd, NONE


;##/


; ==================================================================================================
;     XPT_PropertyNode
; ==================================================================================================

;##\

; 
; Method:    XPT_PropertyNode.Done
; Purpose:   Finalizes the XPT_PropertyNode object and all childs in its branch.
; Arguments: None.
; Return:    Nothing.

Method XPT_PropertyNode.Done, uses esi
    ;DbgText "XPT_PropertyNode.Done"
    SetObject esi
    invoke DeleteObject, [esi].hFont
    ACall esi.Done
MethodEnd, NONE

; 
; Method:    XPT_PropertyNode.Init
; Purpose:   Initializes a XPT_PropertyNode object.
; Arguments: Arg1: -> Owner object, XTreeView.
;            Arg2: Icon Handle.
;            Arg2: -> Node Caption Text.
;            Arg3: Node height.
;            Arg4: Node state (XTVNS_DISABLED, XTVNS_EXPANDED, etc.).
;            Arg5: -> Variable
;            Arg6: Variable type.
;            Arg7: -> Control option list.
; Return:    Nothing.

Method XPT_PropertyNode.Init, uses edi esi, pOwner:Pointer, hIcon:Handle, pCaption:Pointer, \
                                      dHeight:dword, dState:dword, \
                                      pVariable:Pointer, dVarType:dword, pCtrlOptions:Pointer
    local hPrvFont:Handle, TxtMtr:TEXTMETRIC

    SetObject esi
    ACall esi.Init, pOwner, XPT_PROPERTY_NAME_WIDTH, dHeight, dState
    m2m [esi].hIcon, hIcon, eax
    m2m [esi].pCaption, pCaption, eax
    m2m [esi].pVariable, pVariable, eax
    m2m [esi].dVarType, dVarType, eax
    m2m [esi].pCtrlOptions, pCtrlOptions, eax

    ;Create the text font
    mov edi, [esi].pOwner
    assume edi:ptr XPropertyTree
    invoke GetDeviceCaps, [edi].hMemDC, LOGPIXELSY
    invoke MulDiv, XPT_FONT_SIZE_PROP, eax, -72
    mov XPT_LogFontText.lfHeight, eax
    mov XPT_LogFontText.lfWeight, FW_NORMAL
    mov [esi].hFont, $invoke(CreateFontIndirect, offset XPT_LogFontText)

    ;Calculate font metrics
    mov hPrvFont, $invoke(SelectObject, [edi].hMemDC, [esi].hFont)
    invoke GetTextMetrics, [edi].hMemDC, addr TxtMtr
    mov eax, TxtMtr.tmAscent
    add eax, TxtMtr.tmDescent
    mov [esi].dFontHeight, eax
    m2m [esi].dFontAveWidth, TxtMtr.tmAveCharWidth

    ;Housekeeing
    invoke SelectObject, [edi].hMemDC, hPrvFont
    assume edi:NOTHING
MethodEnd, NONE

; 
; Method:    XPT_PropertyNode.CanDragMe
; Purpose:   Indicates if a drag operation can be started.
; Arguments: Arg1: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    Always FALSE => no dragging allowed.

Method XPT_PropertyNode.CanDragMe, NOFRAME, dKeyFlags:dword
    xor eax, eax
MethodEnd 8, eax

; 
; Method:    XPT_PropertyNode.MouseHit
; Purpose:   Checks if the mouse has clicked a hotspot and performs an action.
; Arguments: Arg1: button that triggered this action.
;            Arg2: -> clicked point.
;            Arg3: -> XTVN_DISP_INFO.
;            Arg4: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    eax = TRUE if and action was triggered, otherwise FALSE.

Method XPT_PropertyNode.MouseHit, uses ebx esi, dButton:dword, pPoint:Pointer, pDispInfo:Pointer, \
                                                dKeyFlags:dword
    ;DbgText "XPT_PropertyNode.MouseHit"
    .if pDispInfo != NULL
      SetObject esi
      mov ebx, [esi].pOwner

      assume ebx:ptr XPropertyTree
      .if esi != [ebx].pSelectedNode
        OCall ebx::XTreeView.GetDispInfo, [ebx].pSelectedNode
        push [ebx].pSelectedNode
        push eax
        OCall ebx::XTreeView.Select, esi
        OCall esi.RenderContent, pDispInfo
        pop eax
        pop ecx
        .if eax != NULL
          OCall ecx::XTreeViewNode.RenderContent, eax
        .endif
      .endif
      invoke UpdateWindow, [ebx].hWnd
      mov eax, TRUE

      assume ebx:NOTHING
    .else
      xor eax, eax
    .endif
MethodEnd, eax

; 
; Method:    XPT_PropertyNode.MouseHitDbl
; Purpose:   Checks if the mouse has double clicked a hotspot and performs an action.
; Arguments: Arg1: button that triggered this action.
;            Arg2: -> clicked point.
;            Arg3: -> XTVN_DISP_INFO.
;            Arg4: Key flags (MK_CONTROL, MK_LBUTTON, MK_MBUTTON, MK_RBUTTON, MK_SHIFT).
; Return:    eax = TRUE if and action was triggered, otherwise FALSE.

Method XPT_PropertyNode.MouseHitDbl, uses ebx esi, dButton:dword, pPoint:Pointer, \
                                     pDispInfo:Pointer, dKeyFlags:dword
    local bBuffer[20]:byte, ControlDef:XPT_ControlDef

    ;DbgText "XPT_PropertyNode.MouseHitDbl"
    .if pDispInfo != NULL
      SetObject esi
      .if dButton == MK_LBUTTON
        ;Enter edit mode
        mov ebx, [esi].pOwner
        mov edx, pDispInfo
        assume edx:ptr XPT_DISP_INFO, ebx:ptr XPropertyTree

        BitSet [ebx].dState, XPTS_CHILD_FOCUSED        ;Set BEFORE the control is created!

        .if [esi].dVarType == XPT_VAR_TYPE_STRING
          mov ControlDef.dCtlID, XPT_CONTROL_ID_EDIT
          mov eax, [esi].dFontHeight
          mov ControlDef.dHeight, eax
          mov ecx, [edx].TextRect.bottom
          sub ecx, [edx].TextRect.top
          sub ecx, eax
          shr ecx, 1
          add ecx, [edx].TextRect.top
          mov ControlDef.sdPosY, ecx
          mov eax, [edx].TextRect.right
          sub eax, [edx].TextRect.left
          mov ControlDef.dWidth, eax
          m2m ControlDef.sdPosX, [edx].TextRect.left
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or ES_AUTOHSCROLL
          mov ControlDef.dExStyle, 0
          m2m ControlDef.pTitle, [esi].pVariable
          mov [ebx].pEdtCtrl, $New(XPT_Edit, Init, esi, [ebx].hWnd, addr ControlDef)
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          invoke SendMessage, ebx, EM_SETMARGINS, EC_LEFTMARGIN, [esi].dFontAveWidth
          invoke ReleaseCapture
          invoke SetFocus, ebx

        .elseif [esi].dVarType == XPT_VAR_TYPE_INTEGER
          mov ControlDef.dCtlID, XPT_CONTROL_ID_EDIT
          mov eax, [esi].dFontHeight
          mov ControlDef.dHeight, eax
          mov ecx, [edx].TextRect.bottom
          sub ecx, [edx].TextRect.top
          sub ecx, eax
          shr ecx, 1
          add ecx, [edx].TextRect.top
          mov ControlDef.sdPosY, ecx
          mov eax, [edx].TextRect.right
          sub eax, [edx].TextRect.left
          mov ControlDef.dWidth, eax
          m2m ControlDef.sdPosX, [edx].TextRect.left
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or ES_AUTOHSCROLL OR ES_NUMBER
          mov ControlDef.dExStyle, 0
          lea ecx, bBuffer
          mov ControlDef.pTitle, ecx
          mov eax, [esi].pVariable
          invoke dword2dec, ecx, dword ptr [eax]
          mov [ebx].pEdtCtrl, $New(XPT_Edit, Init, esi, [ebx].hWnd, addr ControlDef)
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          invoke SendMessage, ebx, EM_SETMARGINS, EC_LEFTMARGIN, [esi].dFontAveWidth
          invoke ReleaseCapture
          invoke SetFocus, ebx

        .elseif [esi].dVarType == XPT_VAR_TYPE_REAL4
          mov ControlDef.dCtlID, XPT_CONTROL_ID_EDIT
          mov eax, [esi].dFontHeight
          mov ControlDef.dHeight, eax
          mov ecx, [edx].TextRect.bottom
          sub ecx, [edx].TextRect.top
          sub ecx, eax
          shr ecx, 1
          add ecx, [edx].TextRect.top
          mov ControlDef.sdPosY, ecx
          mov eax, [edx].TextRect.right
          sub eax, [edx].TextRect.left
          mov ControlDef.dWidth, eax
          m2m ControlDef.sdPosX, [edx].TextRect.left
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or ES_AUTOHSCROLL
          mov ControlDef.dExStyle, 0
          lea ecx, bBuffer
          mov ControlDef.pTitle, ecx
          mov eax, [esi].pVariable
          fld real4 ptr [eax]
          invoke St0ToStr, ecx, 0, 8, 0
          fstp st(0)
          mov [ebx].pEdtCtrl, $New(XPT_Edit, Init, esi, [ebx].hWnd, addr ControlDef)
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          invoke SendMessage, ebx, EM_SETMARGINS, EC_LEFTMARGIN, [esi].dFontAveWidth
          invoke ReleaseCapture
          invoke SetFocus, ebx

        .elseif [esi].dVarType == XPT_VAR_TYPE_COLOR
          mov eax, [edx].ButtonRect.right
          sub eax, [edx].ButtonRect.left
          mov ControlDef.dWidth, eax
          mov ecx, [edx].ButtonRect.bottom
          sub ecx, [edx].ButtonRect.top
          mov ControlDef.dHeight, ecx
          m2m ControlDef.sdPosX, [edx].ButtonRect.left
          inc ControlDef.sdPosX
          m2m ControlDef.sdPosY, [edx].ButtonRect.top
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or BS_NOTIFY or BS_OWNERDRAW
          mov ControlDef.dExStyle, 0
          mov ControlDef.pTitle, NULL
          mov [ebx].pBtnCtrl, $New(XPT_Button, Init, esi, [ebx].hWnd, addr ControlDef)
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          invoke ReleaseCapture
          invoke SetFocus, ebx

        .elseif ([esi].dVarType == XPT_VAR_TYPE_FILE) || ([esi].dVarType == XPT_VAR_TYPE_PATH)
          mov ControlDef.dCtlID, XPT_CONTROL_ID_BUTTON
          mov eax, [edx].ButtonRect.right
          sub eax, [edx].ButtonRect.left
          mov ControlDef.dWidth, eax
          mov ecx, [edx].ButtonRect.bottom
          sub ecx, [edx].ButtonRect.top
          mov ControlDef.dHeight, ecx
          m2m ControlDef.sdPosX, [edx].ButtonRect.left
          inc ControlDef.sdPosX
          m2m ControlDef.sdPosY, [edx].ButtonRect.top
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or BS_NOTIFY or BS_OWNERDRAW
          mov ControlDef.dExStyle, 0
          mov ControlDef.pTitle, NULL
          mov [ebx].pBtnCtrl, $New(XPT_Button, Init, esi, [ebx].hWnd, addr ControlDef)
          invoke SendMessage, [eax].WinControl.hWnd, WM_SETFONT, [esi].hFont, FALSE
          mov ControlDef.dCtlID, XPT_CONTROL_ID_EDIT
          mov edx, pDispInfo
          mov eax, [edx].TextRect.right
          sub eax, [edx].TextRect.left
          mov ControlDef.dWidth, eax
          mov eax, [esi].dFontHeight
          mov ControlDef.dHeight, eax
          mov ecx, [edx].TextRect.bottom
          sub ecx, [edx].TextRect.top
          sub ecx, eax
          shr ecx, 1
          add ecx, [edx].TextRect.top
          mov ControlDef.sdPosY, ecx
          m2m ControlDef.sdPosX, [edx].TextRect.left
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or ES_AUTOHSCROLL
          mov ControlDef.dExStyle, 0
          m2m ControlDef.pTitle, [esi].pVariable
          mov [ebx].pEdtCtrl, $New(XPT_Edit, Init, esi, [ebx].hWnd, addr ControlDef)
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          invoke SendMessage, ebx, EM_SETMARGINS, EC_LEFTMARGIN, [esi].dFontAveWidth
          invoke ReleaseCapture
          invoke SetFocus, ebx

        .elseif ([esi].dVarType == XPT_VAR_TYPE_LIST)
          mov ControlDef.dCtlID, XPT_CONTROL_ID_COMBOBOX
          mov edx, pDispInfo
          mov eax, [edx].ButtonRect.right
          sub eax, [edx].TextRect.left
          inc eax
          mov ControlDef.dWidth, eax
          mov ControlDef.dHeight, 100
          mov ecx, [edx].TextRect.top
          mov ControlDef.sdPosY, ecx
          mov eax, [edx].TextRect.left
          m2m ControlDef.sdPosX, eax
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or CBS_DROPDOWNLIST or CBS_OWNERDRAWFIXED
          mov ControlDef.dExStyle, 0
          m2m ControlDef.pTitle, $OfsCStr("ComboBox")
          mov [ebx].pCBxCtrl, $New(XPT_ComboBox)
          OCall [ebx].pCBxCtrl::XPT_ComboBox.Init, esi, [ebx].hWnd, addr ControlDef
          mov eax, [ebx].pCBxCtrl
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          ;Set data
          mov edx, [esi].pCtrlOptions
          .if edx != NULL
            xor eax, eax
            .while (byte ptr [edx] != 0)
              push eax
              push edx
              invoke SendMessage, ebx, CB_ADDSTRING, 0, edx
              invoke StrLength, Pointer ptr [esp]
              pop edx
              pop eax
              add edx, ecx
              inc eax
            .endw
            mov ControlDef.pTitle, NULL
          .endif

          mov ecx, [esi].pVariable
          invoke SendMessage, ebx, CB_SETCURSEL, dword ptr [ecx], 0

          invoke ReleaseCapture
          invoke SetFocus, ebx

        .elseif ([esi].dVarType == XPT_VAR_TYPE_SYS_COLOR)
          mov ControlDef.dCtlID, XPT_CONTROL_ID_COMBOBOX
          mov edx, pDispInfo
          mov eax, [edx].ButtonRect.right
          sub eax, [edx].TextRect.left
          inc eax
          mov ControlDef.dWidth, eax
          mov ControlDef.dHeight, 100
          mov ecx, [edx].TextRect.top
          mov ControlDef.sdPosY, ecx
          mov eax, [edx].TextRect.left
          m2m ControlDef.sdPosX, eax
          mov ControlDef.dStyle, WS_CHILD or WS_VISIBLE or CBS_DROPDOWNLIST or CBS_OWNERDRAWFIXED
          mov ControlDef.dExStyle, 0
          m2m ControlDef.pTitle, $OfsCStr("ComboBox")
          mov [ebx].pCBxCtrl, $New(XPT_ComboBox)
          OCall [ebx].pCBxCtrl::XPT_ComboBox.Init, esi, [ebx].hWnd, addr ControlDef
          mov eax, [ebx].pCBxCtrl
          mov ebx, [eax].WinControl.hWnd
          invoke SendMessage, ebx, WM_SETFONT, [esi].hFont, FALSE
          ;Set data
          mov edx, offset XPT_SYS_COLOR_STR
          .while (byte ptr [edx] != 0)
            push edx
            invoke SendMessage, ebx, CB_ADDSTRING, 0, edx
            invoke StrLength, Pointer ptr [esp]
            pop edx
            add edx, ecx
          .endw
          mov ControlDef.pTitle, NULL

          ;Find the value of the variable in the lookup table
          mov ecx, [esi].pVariable
          mov eax, dword ptr [ecx]
          mov edx, offset XPT_SYS_COLOR_VAL
          .while edx < offset XPT_SYS_COLOR_VAL + XPT_SYS_COLOR_TABLE_SIZE
            movzx ecx, byte ptr [edx]
            .break .if ecx == eax
            inc edx
          .endw
          sub edx, offset XPT_SYS_COLOR_VAL
          .if edx == XPT_SYS_COLOR_TABLE_SIZE
            mov edx, -1
          .endif
          ;Set the found index
          invoke SendMessage, ebx, CB_SETCURSEL, edx, 0

          invoke ReleaseCapture
          invoke SetFocus, ebx

        .endif

      .endif

      assume edx:NOTHING, ebx:NOTHING
      mov eax, TRUE
    .else
      xor eax, eax
    .endif
MethodEnd, eax

; 
; Method:    XPT_PropertyNode.Render
; Purpose:   Draws the complete node.
; Arguments: Arg1: -> XTVN_DISP_INFO.
; Return:    Nothing.

Method XPT_PropertyNode.Render, uses esi, pDispInfo:Pointer
    ;DbgText "XPT_PropertyNode.Render"
    SetObject esi
    .if ([esi].dHeight != 0) || ([esi].dWidth != 0)     ;Don't draw nodes with height and width = 0
      mov edx, [esi].pOwner

      ;Calc the total indentation of this node
      mov ecx, [esi].pParent
      mov eax, [edx].XPropertyTree.ClientOffset.x
      mov edx, [esi].pOwner
      sub eax, [edx].XPropertyTree.TotalOffset.x
      add eax, [edx].XPropertyTree.dNodeIndent
      .while ecx != NULL
        mov ecx, [ecx].XTreeViewNode.pParent
        add eax, [edx].XPropertyTree.dNodeIndent
      .endw                                             ;eax = total indentation

      ;Fill XTVN_DISP_INFO.ContentRect info
      mov ecx, pDispInfo
      mov [ecx].XTVN_DISP_INFO.ContentRect.left, eax
      m2m [ecx].XTVN_DISP_INFO.ContentRect.right, [edx].XPropertyTree.ClientSize.x, eax
      m2m [ecx].XTVN_DISP_INFO.ContentRect.top, [ecx].XTVN_DISP_INFO.sdNodeTop, eax
      m2m [ecx].XTVN_DISP_INFO.ContentRect.bottom, [ecx].XTVN_DISP_INFO.sdNodeBot, eax

      if DEBUGGING
      .if sdword ptr [ecx].XTVN_DISP_INFO.ContentRect.bottom < 0
        DbgWarning "Error while positioning the view"
      .endif
      endif

      ;Render the tree structure (lines and toggler)
      OCall esi.RenderTreeStruc, pDispInfo

      ;Render the content of the node
      OCall esi.RenderContent, pDispInfo

    .endif
MethodEnd, NONE

; 
; Method:    XPT_PropertyNode.RenderContent
; Purpose:   Draws the content (icon + caption + text) of the node.
; Arguments: Arg1: -> XPT_DISP_INFO
; Return:    Nothing.
; Note:      This method must update the coords of all Rect structures in the Content and must
;            invalidate the modified region.

Method XPT_PropertyNode.RenderContent, uses ebx edi esi, pDispInfo:Pointer
    local dPrevColor:dword, dPrevBkMode:dword, hFontPrev:Handle
    local hPenPrev:Handle, hBrush:Handle, dBot:dword, dTop:dword
    local bBuffer[MAX_PATH]:byte, pText:Pointer, OutputRect:RECT
    local RRect:RECT

    ;DbgText "XPT_PropertyNode.RenderContent"
    mov edi, pDispInfo
    .if edi != NULL
      SetObject esi
      mov ebx, [esi].pOwner
      assume edi:ptr XPT_DISP_INFO, ebx:ptr XPropertyTree
      m2m [edi].TextRect.left, [edi].ContentRect.left, eax
      m2m [edi].TextRect.right, [edi].ContentRect.right, eax

      ;Render the background
      invoke CreateSolidBrush, [ebx].dColorBackground
      push eax
      push eax
      lea eax, [edi].ContentRect
      push eax
      push [ebx].hMemDC
      call FillRect
      call DeleteObject

      ;Calc all rectangles
      mov eax, [edi].ContentRect.bottom
      dec eax                                         ;Make room for the bottom separation line
      mov [edi].CaptionRect.bottom, eax
      mov [edi].TextRect.bottom, eax
      mov edx, [edi].ContentRect.top
      mov [edi].CaptionRect.top, edx
      mov [edi].TextRect.top, edx
      sub eax, edx
      sub eax, XPT_ICON_SIZE
      shr eax, 1
      mov ecx, [edi].CaptionRect.bottom
      sub ecx, eax
      mov [edi].IconRect.bottom, ecx                  ;IconRect.bottom
      sub ecx, XPT_ICON_SIZE
      mov [edi].IconRect.top, ecx                     ;IconRect.top

      mov edx, [edi].ContentRect.left
      add edx, XPT_MARGIN
      mov [edi].IconRect.left, edx                    ;IconRect.left
      mov eax, edx
      add eax, XPT_ICON_SIZE + XPT_MARGIN
      mov [edi].CaptionRect.left, eax                 ;CaptionRect.left
      mov [edi].IconRect.right, eax                   ;IconRect.right
      add eax, [esi].dWidth
      mov [edi].CaptionRect.right, eax                ;CaptionRect.right
      inc eax                                         ;Make room for the vertical separation line
      mov [edi].TextRect.left, eax                    ;TextRect.left
      mov eax, [ebx].ClientSize.x
      mov [edi].ButtonRect.right, eax                ;ButtonRect.right
      .if ([esi].dVarType != XPT_VAR_TYPE_STRING) && \
          ([esi].dVarType != XPT_VAR_TYPE_INTEGER) && \
          ([esi].dVarType != XPT_VAR_TYPE_REAL4)
        sub eax, [edi].CaptionRect.bottom
        add eax, [edi].CaptionRect.top
      .endif
      mov [edi].TextRect.right, eax                   ;TextRect.right
      mov [edi].ButtonRect.left, eax                  ;ButtonRect.left

      m2m [edi].ButtonRect.bottom, [edi].TextRect.bottom         ;ButtonRect.bottom
      m2m [edi].ButtonRect.top, [edi].TextRect.top               ;ButtonRect.top

      ;Render the icon
      invoke DrawIconEx, [ebx].hMemDC, [edi].IconRect.left, [edi].IconRect.top, [esi].hIcon, \
                         XPT_ICON_SIZE, XPT_ICON_SIZE, NULL, NULL, DI_NORMAL

      ;Set text and background colors
      ;  Check if the whole XTV or the single node is disabled
      .if ([ebx].dState & XTVS_DISABLED) || ([esi].dState & XTVNS_DISABLED)
        ;Disabled
        invoke GetSysColor, COLOR_BTNHILIGHT
        mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
        mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
      .else
        ;Enabled
        .if esi == [ebx].pSelectedNode
          .if ([ebx].dState & XTVS_FOCUSED) || ([ebx].dState & XPTS_CHILD_FOCUSED)
            .ifBitSet [esi].dState, XTVNS_HOTTRACK      ;Check if the node is beeing hottracked
              mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, $RGB(255,255,0))
            .else
              invoke GetSysColor, COLOR_HIGHLIGHTTEXT
              mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
            .endif
            invoke GetSysColor, COLOR_HIGHLIGHT
            mov hBrush, $invoke(CreateSolidBrush, eax)
          .else
            invoke GetSysColor, COLOR_WINDOWTEXT
            mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
            invoke GetSysColor, COLOR_BTNFACE
            mov hBrush, $invoke(CreateSolidBrush, eax)
          .endif
        .elseIfBitSet [esi].dState, XTVNS_HOTTRACK      ;Check if the node is beeing hottracked
          mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, $RGB(0,0,255))
          mov hBrush, $invoke(CreateSolidBrush, [ebx].dColorBackground)
        .else
          invoke GetSysColor, COLOR_WINDOWTEXT
          mov dPrevColor, $invoke(SetTextColor, [ebx].hMemDC, eax)
          mov hBrush, $invoke(CreateSolidBrush, [ebx].dColorBackground)
        .endif
      .endif


      ;Set the font
      mov hFontPrev, $invoke(SelectObject, [ebx].hMemDC, [esi].hFont)

      ;Draw the property caption
      mov dPrevBkMode, $invoke(SetBkMode, [ebx].hMemDC, TRANSPARENT)
      invoke FillRect, [ebx].hMemDC, addr [edi].CaptionRect, hBrush
      invoke DeleteObject, hBrush

      mov eax, [edi].CaptionRect.bottom
      sub eax, [edi].CaptionRect.top
      sub eax, [esi].dFontHeight
      shr eax, 1
      add eax, [edi].CaptionRect.top
      mov OutputRect.top, eax
      add eax, [esi].dFontHeight
      mov OutputRect.bottom, eax

      mov ecx, [edi].CaptionRect.left
      add ecx, [esi].dFontAveWidth
      mov OutputRect.left, ecx
      mov ecx, [edi].CaptionRect.right
      sub ecx, [esi].dFontAveWidth
      mov OutputRect.right, ecx

      invoke DrawText, [ebx].hMemDC, [esi].pCaption , -1, addr OutputRect, \
                       DT_SINGLELINE or DT_LEFT or DT_END_ELLIPSIS

      ;Draw the property text
      .if (esi == [ebx].pSelectedNode) && (([ebx].pEdtCtrl != NULL) || ([ebx].pBtnCtrl != NULL))
        push $invoke(CreateSolidBrush, $invoke(GetSysColor, COLOR_WINDOW))
        invoke FillRect, [ebx].hMemDC, addr [edi].TextRect, eax
        call DeleteObject
      .endif

      ;Draw a colored rect in case of XPT_VAR_TYPE_COLOR or XPT_VAR_TYPE_SYS_COLOR
      .if [esi].dVarType == XPT_VAR_TYPE_COLOR
        mov ecx, [esi].pVariable
        invoke CreateSolidBrush, dword ptr [ecx]
        push $invoke(SelectObject, [ebx].hMemDC, eax)
        push [ebx].hMemDC

        push [edi].IconRect.bottom
        mov eax, [edi].TextRect.left
        add eax, XPT_MARGIN + XPT_ICON_SIZE
        push eax
        push [edi].IconRect.top
        sub eax, XPT_ICON_SIZE
        push eax
        push [ebx].hMemDC
        call Rectangle
        call SelectObject
        invoke DeleteObject, eax

      .elseif [esi].dVarType == XPT_VAR_TYPE_SYS_COLOR
        ;Draw a rectangle with the corresponding system color
        mov ecx, [esi].pVariable
        mov hBrush, $invoke(GetSysColorBrush, dword ptr [ecx])  ;Don't delete this brush
        .if eax == 0
          invoke GetStockObject, HOLLOW_BRUSH
        .endif
        push $invoke(SelectObject, [ebx].hMemDC, eax)   ;Create stack for SelectObject API
        push [ebx].hMemDC

        mov eax, [edi].IconRect.bottom                  ;Create stack for Rectange API
        mov RRect.bottom, eax
        push eax
        mov eax, [edi].TextRect.left
        add eax, XPT_MARGIN + XPT_ICON_SIZE
        push eax
        mov RRect.right, eax
        mov ecx, [edi].IconRect.top
        mov RRect.top, ecx
        push ecx
        sub eax, XPT_ICON_SIZE
        mov RRect.left, eax
        push eax
        push [ebx].hMemDC
        call Rectangle
        call SelectObject

        ;In case the variable has an invalid system color index, then draw a cross
        .if hBrush == 0
          invoke MoveToEx, [ebx].hMemDC, RRect.left, RRect.top, NULL
          mov eax, RRect.bottom
          dec eax
          push eax
          mov ecx, RRect.right
          dec ecx
          push ecx
          push [ebx].hMemDC
          call LineTo
          push NULL
          push RRect.top
          mov eax, RRect.right
          dec eax
          push eax
          push [ebx].hMemDC
          call MoveToEx
          mov eax, RRect.bottom
          dec eax
          push eax
          push RRect.left
          push [ebx].hMemDC
          call LineTo
        .endif
      .endif

      ;Draw the text of the variable
      mov ecx, [edi].TextRect.left
      add ecx, [esi].dFontAveWidth
      mov OutputRect.left, ecx
      .if [ebx].pEdtCtrl == NULL
        mov ecx, [edi].ButtonRect.right
      .else
        mov ecx, [edi].TextRect.right
      .endif
      sub ecx, [esi].dFontAveWidth
      mov OutputRect.right, ecx

      invoke SetTextColor, [ebx].hMemDC, $RGB(0,0,0)

      ;Get the correct string
      .if [esi].dVarType == XPT_VAR_TYPE_STRING
        m2m pText, [esi].pVariable
      .elseif [esi].dVarType == XPT_VAR_TYPE_INTEGER
        lea ecx, bBuffer
        mov pText, ecx
        mov eax, [esi].pVariable
        invoke dword2dec, ecx, dword ptr [eax]
      .elseif [esi].dVarType == XPT_VAR_TYPE_REAL4
        lea ecx, bBuffer
        mov pText, ecx
        mov eax, [esi].pVariable
        fld real4 ptr [eax]
        invoke St0ToStr, ecx, 0, 8, 0
        fstp st(0)
      .elseif [esi].dVarType == XPT_VAR_TYPE_COLOR
        add OutputRect.left, 2 * XPT_MARGIN + XPT_ICON_SIZE
        .ifBitSet [ebx].dState, XPTS_CHILD_FOCUSED
          mov ecx, [edi].ButtonRect.right
          sub ecx, [edi].ButtonRect.left
          sub OutputRect.right, ecx
        .endif
        lea ecx, bBuffer
        mov pText, ecx
        mov eax, [esi].pVariable
        invoke dword2hex, ecx, dword ptr [eax]
      .elseif [esi].dVarType == XPT_VAR_TYPE_FILE
        m2m pText, [esi].pVariable
      .elseif [esi].dVarType == XPT_VAR_TYPE_PATH
        m2m pText, [esi].pVariable
      .elseif [esi].dVarType == XPT_VAR_TYPE_LIST
        mov edx, [esi].pCtrlOptions
        xor eax, eax
        mov pText, eax
        .if edx != NULL
          .while (byte ptr [edx] != 0)
            mov ecx, [esi].pVariable
            .break .if eax == dword ptr [ecx]
            push eax
            push edx
            invoke StrLength, edx
            pop edx
            pop eax
            add edx, ecx
            inc eax
          .endw
          mov pText, edx
        .endif
      .elseif [esi].dVarType == XPT_VAR_TYPE_SYS_COLOR
        add OutputRect.left, 2 * XPT_MARGIN + XPT_ICON_SIZE
        .ifBitSet [ebx].dState, XPTS_CHILD_FOCUSED      ;If it has the focus,
          mov ecx, [edi].ButtonRect.right               ;  then make room for the button
          sub ecx, [edi].ButtonRect.left
          sub OutputRect.right, ecx
        .endif
        push esi
        push edi
        push ebx
        mov edi, offset XPT_SYS_COLOR_STR
        mov ebx, offset XPT_SYS_COLOR_VAL
        mov eax, [esi].pVariable
        mov esi, dword ptr [eax]
        mov pText, NULL
        .while byte ptr [edi] != 0
          movzx ecx, byte ptr [ebx]
          .break .if ecx == esi
          invoke StrLength, edi
          add edi, ecx
          inc ebx
        .endw
        .if byte ptr [edi] != 0
          mov pText, edi
        .else
          mov pText, offset XPT_SYS_COLOR_STR_ERROR
        .endif
        pop ebx
        pop edi
        pop esi
      .else
        mov pText, NULL
      .endif

      invoke DrawText, [ebx].hMemDC, pText , -1, addr OutputRect, \
                       DT_SINGLELINE or DT_LEFT or DT_END_ELLIPSIS

      ;Restore previous settings
      invoke SelectObject, [ebx].hMemDC, hFontPrev
      invoke SetBkMode, [ebx].hMemDC, dPrevBkMode
      invoke SetTextColor, [ebx].hMemDC, dPrevColor

      ;Render some separation lines
      mov hPenPrev, $invoke(SelectObject, [ebx].hMemDC, [ebx].hPenLine)
      mov eax, [edi].ContentRect.bottom
      dec eax
      mov dBot, eax
      mov eax, [edi].ContentRect.top
      dec eax
      mov dTop, eax
      invoke MoveToEx, [ebx].hMemDC, [edi].ContentRect.left, dBot, NULL
      invoke LineTo, [ebx].hMemDC, [ebx].ClientSize.x, dBot
      invoke MoveToEx, [ebx].hMemDC, [edi].CaptionRect.right, dBot, NULL
      invoke LineTo, [ebx].hMemDC, [edi].CaptionRect.right, dTop
      invoke SelectObject, [ebx].hMemDC, hPenPrev

      ;Invalidate the new drawn rect
      invoke InvalidateRect, [ebx].hWnd, addr [edi].ContentRect, FALSE

      assume edi:NOTHING, ebx:NOTHING
    .endif
MethodEnd, NONE


;##/

; ==================================================================================================
;     XPropertyTree
; ==================================================================================================

;##\

; 
; Method:    XPropertyTree.Draw
; Purpose:   Renders all nodes in view on the back buffer.
; Arguments: None.
; Return:    Nothing.

Method XPropertyTree.Draw, uses ebx edi esi
    local Rct:RECT, ScrollInfo:SCROLLINFO, RemRect:RECT

    ;DbgText "XPropertyTree.Draw"
    SetObject esi
    .ifBitClr [esi].dState, XTVS_LOCKED

      ;Some initialization
      OCall [esi].DispInfoColl::DwordCollection.DisposeAll      ;Release all XTVN_DISP_INFO structs

      ;(De)acticate the vertical scrollbar
      xor ecx, ecx
      mov eax, [esi].TotalSize.y
      sub eax, [esi].ClientSize.y
      setns cl
      invoke ShowScrollBar, [esi].hWnd, SB_VERT, ecx
      invoke GetClientRect, [esi].hWnd, addr Rct
      m2m [esi].ClientSize.x, Rct.right, eax
      m2m [esi].ClientSize.y, Rct.bottom, eax

      ;Manage the vertical scrollbar
      mov eax, [esi].TotalSize.y
      .if eax > [esi].ClientSize.y
        mov ScrollInfo.SCROLLINFO.cbSize, sizeof SCROLLINFO
        mov ScrollInfo.SCROLLINFO.fMask, SIF_POS or SIF_PAGE

        ;Scale the values to fit in a 16 bit range
        xor edx, edx
        mov eax, [esi].TotalOffset.y
        mov ecx, 0FFFFh
        mul ecx
        div [esi].TotalSize.y
        mov ScrollInfo.SCROLLINFO.nPos, eax

        mov eax, [esi].ClientSize.y
        mov ecx, 0FFFFh
        mul ecx
        div [esi].TotalSize.y
        .if eax == 0
          inc eax
        .endif
        mov ScrollInfo.SCROLLINFO.nPage, eax

        invoke SetScrollInfo, [esi].hWnd, SB_VERT, addr ScrollInfo, TRUE
      .endif

      ;Draw all nodes in view
      .if [esi].pFirstVisibleNode == NULL
        OCall esi.Set1stVisibleNode, [esi].pFirstChild
        mov [esi].TotalOffset.x, 0
        mov [esi].TotalOffset.y, 0
      .endif

      mov edi, [esi].dFirstOffset
      neg edi
      mov ebx, [esi].pFirstVisibleNode
      .if ebx != NULL
        .repeat
          OCall [esi].DispInfoPool::DataPool.NewItem
          mov [eax].XTVN_DISP_INFO.pNode, ebx
          mov [eax].XTVN_DISP_INFO.sdNodeTop, edi
          add edi, [ebx].XTreeViewNode.dHeight
          mov [eax].XTVN_DISP_INFO.sdNodeBot, edi
          OCall [esi].DispInfoColl::DwordCollection.Insert, eax
          OCall ebx::XTreeViewNode.Render, eax
          XTV_BrowseFrwd ebx
          .break .if ebx == NULL
        .until edi >= [esi].ClientSize.y
      .endif

      .if edi < [esi].ClientSize.y
        ;Fill the bottom remaining background
        mov RemRect.left, 0
        m2m RemRect.top, edi
        m2m RemRect.right, [esi].ClientSize.x
        m2m RemRect.bottom, [esi].ClientSize.y
        OCall esi.RenderBackground, addr RemRect
        invoke InvalidateRect, [esi].hWnd, addr RemRect, FALSE
      .endif
    .endif
MethodEnd, NONE

; 
; Method:    XPropertyTree.Init
; Purpose:   Initializes a XTreeView object.
; Arguments: Arg1: -> owner object.
;            Arg2: Parent handle.
;            Arg2: -> Definition structure.
; Return:    Nothing.

Method XPropertyTree.Init, uses esi, pOwner:Pointer, hParentWnd:Handle, pDefStruc:Pointer
    SetObject esi
    mov ecx, pDefStruc
    assume ecx:ptr XTV_DEF
    mov eax, [ecx].dStyle
    or eax, WS_CHILD or WS_VISIBLE or WS_TABSTOP; or WS_CLIPCHILDREN or WS_CLIPSIBLINGS
    invoke CreateWindowEx, [ecx].dExStyle, offset szXPT_Class, [ecx].pTitle, \
           eax, [ecx].sdPosX, [ecx].sdPosY, [ecx].dWidth, [ecx].dHeight, \
           hParentWnd, [ecx].dCtlID, hInstance, esi
    assume ecx:NOTHING
    TCall esi::WinPrimer.Init, pOwner, [esi].hWnd

MethodEnd, NONE

; 
; Method:    XPropertyTree.OnCreate
; Purpose:   Event handler for WM_CREATE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    eax = Zero to continue creation of the window, else -1.

Method XPropertyTree.OnCreate, uses esi, wParam:dword, lParam:dword
    local ScrollInfo:SCROLLINFO

    ;DbgText "XPropertyTree.OnCreate"
    SetObject esi

    ;Some initialization
    mov [esi].hCursorArrow, $invoke(LoadCursor, hInstance, "CURSOR_ARROW")

    OCall [esi].DispInfoColl::DwordCollection.Init, esi, 20, 10, COLLECTIONMAXSIZE
    OCall [esi].DispInfoPool::DataPool.Init, esi, [esi].dDispInfoSize, 4096
    lea ecx, [esi].DispInfoColl
    Override ecx::DwordCollection.DestroyItem, DisplayInfoColl.DestroyItem

    ;Create pens to draw
    mov [esi].hPenLine, $invoke(CreatePen, PS_SOLID, 0, [esi].dColorLine)
    mov [esi].hPenToggler, $invoke(CreatePen, PS_SOLID, 0, [esi].dColorToggler)

    ;Set scroller infos
    mov ScrollInfo.cbSize, sizeof SCROLLINFO
    mov ScrollInfo.fMask, SIF_RANGE
    mov ScrollInfo.nMin, 0
    mov ScrollInfo.nMax, 0FFFFh       ;16 bit max value passed in WM_VSCROLL
    invoke SetScrollInfo, [esi].hWnd, SB_VERT, addr ScrollInfo, FALSE

    xor eax, eax
MethodEnd, eax

; 
; Method:    XPropertyTree.OnCommand
; Purpose:   Event handler for WM_COMMAND message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

; This callback proc sets the initial path for SHBrowseForFolder
XPT_BrowseFolderCallback proc hWnd:Handle, Msg:dword, lParam:dword, pData:dword
    .if Msg == BFFM_INITIALIZED
      invoke SendMessage, hWnd, BFFM_SETSELECTION, TRUE, pData
    .endif
    ret
XPT_BrowseFolderCallback endp

Method XPropertyTree.OnCommand, uses ebx edi esi, wParam:dword, lParam:dword
    SetObject esi
    mov eax, [esi].pEdtCtrl
    or eax, [esi].pBtnCtrl
    or eax, [esi].pCBxCtrl
    and eax, [esi].pSelectedNode
    .if !Zero?
      mov ebx, [esi].pSelectedNode
      mov eax, wParam
      shr eax, 16

      .if ax == EN_KILLFOCUS
        mov eax, [esi].pEdtCtrl
        .if eax != NULL
          mov ecx, [eax].WinControl.hWnd
          .if ecx == lParam
            invoke GetFocus
            mov ecx, [esi].pBtnCtrl
            .if (ecx == NULL) || (eax != [ecx].WinControl.hWnd)
              mov ecx, [esi].pCBxCtrl
              .if (ecx == NULL) || (eax != [ecx].WinControl.hWnd)
                Kill [esi].pEdtCtrl
                mov [esi].pEdtCtrl, NULL
                Destroy [esi].pBtnCtrl
                mov [esi].pBtnCtrl, NULL
                Destroy [esi].pCBxCtrl
                mov [esi].pCBxCtrl, NULL
                BitClr [esi].dState, XPTS_CHILD_FOCUSED
                OCall esi.GetDispInfo, ebx
                OCall ebx::XPT_PropertyNode.RenderContent, eax
                invoke UpdateWindow, [esi].hWnd
                xor eax, eax
                ExitMethod
              .endif
            .endif
          .endif
        .endif

      .elseif ax == BN_KILLFOCUS
        mov eax, [esi].pBtnCtrl
        .if eax != NULL
          mov ecx, [eax].WinControl.hWnd
          .if ecx == lParam
            invoke GetFocus
            mov ecx, [esi].pEdtCtrl
            .if (ecx == NULL) || (eax != [ecx].WinControl.hWnd)
              mov ecx, [esi].pCBxCtrl
              .if (ecx == NULL) || (eax != [ecx].WinControl.hWnd)
                Destroy [esi].pEdtCtrl
                mov [esi].pEdtCtrl, NULL
                Kill [esi].pBtnCtrl
                mov [esi].pBtnCtrl, NULL
                Destroy [esi].pCBxCtrl
                mov [esi].pCBxCtrl, NULL
                BitClr [esi].dState, XPTS_CHILD_FOCUSED
                OCall esi.GetDispInfo, ebx
                OCall ebx::XPT_PropertyNode.RenderContent, eax
                invoke UpdateWindow, [esi].hWnd
                xor eax, eax
                ExitMethod
              .endif
            .endif
          .endif
        .endif

      .elseif ax == CBN_KILLFOCUS
        mov eax, [esi].pCBxCtrl
        .if eax != NULL
          mov ecx, [eax].WinControl.hWnd
          .if ecx == lParam
            invoke GetFocus
            mov ecx, [esi].pEdtCtrl
            .if (ecx == NULL) || (eax != [ecx].WinControl.hWnd)
              mov ecx, [esi].pBtnCtrl
              .if (ecx == NULL) || (eax != [ecx].WinControl.hWnd)
                Destroy [esi].pEdtCtrl
                mov [esi].pEdtCtrl, NULL
                Destroy [esi].pBtnCtrl
                mov [esi].pBtnCtrl, NULL
                mov ecx, [esi].pCBxCtrl
                .ifBitClr [ecx].XPT_ComboBox.dState, XPTCB_ESCAPING
                  invoke SendMessage, lParam, CB_GETCURSEL, 0, 0
                  .if eax != CB_ERR
                    mov ecx, [ebx].XPT_PropertyNode.pVariable
                    add eax, offset XPT_SYS_COLOR_VAL
                    movzx edx, byte ptr [eax]
                    mov dword ptr [ecx], edx
                  .endif
                .endif
                Kill [esi].pCBxCtrl
                mov [esi].pCBxCtrl, NULL
                BitClr [esi].dState, XPTS_CHILD_FOCUSED
                OCall esi.GetDispInfo, ebx
                OCall ebx::XPT_PropertyNode.RenderContent, eax
                invoke UpdateWindow, [esi].hWnd
                xor eax, eax
                ExitMethod
              .endif
            .endif
          .endif
        .endif

      .elseif ax == BN_CLICKED
        mov eax, [esi].pBtnCtrl
        .if eax != NULL
          mov ecx, [eax].WinControl.hWnd
          .if ecx == lParam
            .if [ebx].XPT_PropertyNode.pVariable != NULL
              mov edi, [ebx].XPT_PropertyNode.pCtrlOptions

              .if [ebx].XPT_PropertyNode.dVarType == XPT_VAR_TYPE_COLOR
                .if edi != NULL
                  mov [edi].CHOOSECOLOR.lStructSize, sizeof CHOOSECOLOR
                  m2m [edi].CHOOSECOLOR.hwndOwner, [esi].hWnd
                  mov ecx, [ebx].XPT_PropertyNode.pVariable
                  m2m [edi].CHOOSECOLOR.rgbResult, dword ptr [ecx]
                  invoke ChooseColor, edi
                  .if eax
                    mov ecx, [ebx].XPT_PropertyNode.pVariable
                    m2m dword ptr [ecx], [edi].CHOOSECOLOR.rgbResult
                  .endif
                  OCall esi.GetDispInfo, ebx
                  OCall ebx::XPT_PropertyNode.RenderContent, eax
                  invoke UpdateWindow, [esi].hWnd
                  xor eax, eax
                  ExitMethod
                .endif

              .elseif [ebx].XPT_PropertyNode.dVarType == XPT_VAR_TYPE_FILE
                .if edi != NULL
                  mov [edi].OPENFILENAME.lStructSize, sizeof OPENFILENAME
                  m2m [edi].OPENFILENAME.hwndOwner, [esi].hWnd
                  invoke GetOpenFileName, edi
                  .if eax == TRUE
                    invoke StrCopy, [ebx].XPT_PropertyNode.pVariable, [edi].OPENFILENAME.lpstrFile
                    OCall esi.GetDispInfo, ebx
                    OCall ebx::XPT_PropertyNode.RenderContent, eax
                    invoke UpdateWindow, [esi].hWnd
                    xor eax, eax
                    ExitMethod
                  .endif
                .endif

              .elseif [ebx].XPT_PropertyNode.dVarType == XPT_VAR_TYPE_PATH
                .if edi != NULL
                  m2m [edi].BROWSEINFO.hwndOwner, [esi].hWnd
                  mov [edi].BROWSEINFO.pszDisplayName, NULL
                  mov [edi].BROWSEINFO.lpfn, offset XPT_BrowseFolderCallback
                  m2m [edi].BROWSEINFO.lParam, [ebx].XPT_PropertyNode.pVariable
                  invoke SHBrowseForFolder, edi
                  .if eax != NULL
                    push eax
                    invoke SHGetPathFromIDList, eax, [ebx].XPT_PropertyNode.pVariable
                    call CoTaskMemFree
                    OCall esi.GetDispInfo, ebx
                    OCall ebx::XPT_PropertyNode.RenderContent, eax
                    invoke UpdateWindow, [esi].hWnd
                    xor eax, eax
                    ExitMethod
                  .endif
                .endif
              .endif
            .endif
          .endif
        .endif
      .endif
    .endif
    mov eax, 1
MethodEnd, eax

; 
; Method:    XPropertyTree.OnDestroy
; Purpose:   Event handler for WM_DESTROY message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XPropertyTree.OnDestroy, uses esi, wParam:dword, lParam:dword
    ;DbgText "XPropertyTree.OnDestroy"

    SetObject esi

    OCall [esi].DispInfoColl::DwordCollection.Done
    OCall [esi].DispInfoPool::DataPool.Done

    invoke DeleteObject, [esi].hPenLine
    invoke DeleteObject, [esi].hPenToggler

    invoke DeleteObject, $invoke(SelectObject, [esi].hMemDC, [esi].hBmpPrv)
    invoke DeleteDC, [esi].hMemDC

    xor eax, eax
MethodEnd, eax

; 
; Method:    XPropertyTree.OnDrawItem
; Purpose:   Event handler for WM_DRAWITEM message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    TRUE if handled.

Method XPropertyTree.OnDrawItem, uses ebx edi esi, wParam:dword, lParam:dword
    local OutputRect:RECT, TextRect:RECT, pDispInfo:Pointer
    local hFontPrev:Handle, dPrevBkMode:dword, hWnd:Handle, dPrevColor:dword
    local hPenPrev:Handle, dLeft:dword, hPrevBrush:Handle
    local RRect:RECT

    SetObject esi
    ;DbgText "XPropertyTree.OnDrawItem"
    mov edi, lParam
    .if [esi].pCBxCtrl != NULL
      mov ecx, [esi].pCBxCtrl
      m2m hWnd, [ecx].WinControl.hWnd
      mov ebx, [ecx].WinControl.pOwner                    ;ebx -> XPT_PropertyNode
      mov pDispInfo, $OCall(esi.GetDispInfo, ebx)
      assume ebx:ptr XPT_PropertyNode, edi:ptr DRAWITEMSTRUCT
      mov ecx, [edi].itemState
      .ifBitSet [edi].itemState, ODS_COMBOBOXEDIT
        ;This is the Edit control
        .if [edi].itemID != -1
          mov ecx, [eax].XPT_DISP_INFO.TextRect.bottom
          sub ecx, [eax].XPT_DISP_INFO.TextRect.top
          mov OutputRect.bottom, ecx
          push ecx
          mov ecx, [eax].XPT_DISP_INFO.ContentRect.right
          sub ecx, [eax].XPT_DISP_INFO.TextRect.left
          mov OutputRect.right, ecx
          push ecx
          xor eax, eax
          mov OutputRect.top, eax
          mov OutputRect.left, eax
          push eax
          push eax
          call CreateRectRgn
          push eax
          invoke SelectClipRgn, [edi].hdc, eax
          call DeleteObject
          invoke CreateSolidBrush, $invoke(GetSysColor, COLOR_WINDOW)
          push eax
          invoke FillRect, [edi].hdc, addr OutputRect, eax
          call DeleteObject

          mov dPrevColor, $invoke(SetTextColor, [edi].hdc, $RGB(0,0,0))
          mov hFontPrev, $invoke(SelectObject, [edi].hdc, [ebx].hFont)
          mov dPrevBkMode, $invoke(SetBkMode, [edi].hdc, TRANSPARENT)

          ;In case of XPT_VAR_TYPE_SYS_COLOR, draw a colored rectangle
          .if [ebx].dVarType == XPT_VAR_TYPE_SYS_COLOR
            .if [edi].itemData != -1
              ;Find the matching color index
              push ebx
              push esi
              mov ebx, offset XPT_SYS_COLOR_STR
              mov esi, offset XPT_SYS_COLOR_VAL
              .while byte ptr [ebx] != 0
                .break .if ebx == [edi].itemData          ;Compare pointers
                invoke StrLength, ebx
                add ebx, ecx
                inc esi
              .endw
              movzx eax, byte ptr [esi]                   ;Get the color index
              pop esi
              pop ebx
              invoke GetSysColorBrush, eax
            .else
              invoke GetStockObject, HOLLOW_BRUSH
            .endif
            mov hPrevBrush, $invoke(SelectObject, [edi].hdc, eax)

            ;Render a colored rectangle
            mov eax, OutputRect.bottom                    ;Create the stack for the Rectangle API
            mov ecx, eax
            sub eax, XPT_ICON_SIZE
            shr eax, 1
            sub ecx, eax
            push ecx
            mov RRect.bottom, ecx
            mov eax, XPT_MARGIN + XPT_ICON_SIZE
            push eax
            mov RRect.right, eax
            sub ecx, XPT_ICON_SIZE
            push ecx
            mov RRect.top, ecx
            sub eax, XPT_ICON_SIZE
            push eax
            mov RRect.left, eax
            push [edi].hdc
            call Rectangle
            .if [edi].itemData == -1                      ;If the index is invalid, then draw a cross
              invoke MoveToEx, [edi].hdc, RRect.left, RRect.top, NULL
              mov eax, RRect.bottom
              dec eax
              push eax
              mov ecx, RRect.right
              dec ecx
              push ecx
              push [edi].hdc
              call LineTo
              push NULL
              push RRect.top
              mov eax, RRect.right
              dec eax
              push eax
              push [edi].hdc
              call MoveToEx
              mov eax, RRect.bottom
              dec eax
              push eax
              push RRect.left
              push [edi].hdc
              call LineTo
            .endif
            invoke SelectObject, [edi].hdc, hPrevBrush

            mov TextRect.left, XPT_ICON_SIZE + 2*XPT_MARGIN
          .else
            mov TextRect.left, 0
          .endif

          ;Draw the text
          mov eax, OutputRect.bottom
          sub eax, [ebx].dFontHeight
          shr eax, 1
          mov TextRect.top, eax
          add eax, [ebx].dFontHeight
          mov TextRect.bottom, eax

          mov ecx, [ebx].dFontAveWidth
          add TextRect.left, ecx
          mov ecx, OutputRect.right
          mov eax, pDispInfo
          sub ecx, [eax].XPT_DISP_INFO.ButtonRect.right
          add ecx, [eax].XPT_DISP_INFO.ButtonRect.left
          mov TextRect.right, ecx

          .if [edi].itemData == -1
            mov ecx, offset XPT_SYS_COLOR_STR_ERROR
          .else
            mov ecx, [edi].itemData
          .endif
          invoke DrawText, [edi].hdc, ecx, -1, addr TextRect, \
                           DT_SINGLELINE or DT_LEFT or DT_END_ELLIPSIS

          ;Restore previous settings
          invoke SelectObject, [edi].hdc, hFontPrev
          invoke SetBkMode, [edi].hdc, dPrevBkMode
          invoke SetTextColor, [edi].hdc, dPrevColor

          ;Render a down pointing arrow
          mov eax, pDispInfo
          mov ecx, [eax].XPT_DISP_INFO.ButtonRect.left
          sub ecx, [eax].XPT_DISP_INFO.TextRect.left
          add ecx, 4
          mov OutputRect.left, ecx
          mov ecx, [eax].XPT_DISP_INFO.ButtonRect.right
          sub ecx, [eax].XPT_DISP_INFO.TextRect.left
          sub ecx, 2
          mov OutputRect.right, ecx
          invoke DrawFrameControl, [edi].hdc, addr OutputRect, DFC_MENU, 16 or DFCS_TRANSPARENT

          ;Draw a vertical separation line
          invoke CreatePen, PS_SOLID, 1, [esi].dColorBackground
          push eax
          mov hPenPrev, $invoke(SelectObject, [edi].hdc, eax)
          mov ecx, pDispInfo
          mov eax, [ecx].XPT_DISP_INFO.ButtonRect.left
          sub eax, [ecx].XPT_DISP_INFO.TextRect.left
          mov dLeft, eax
          mov eax, [ecx].XPT_DISP_INFO.ContentRect.bottom
          sub eax, [ecx].XPT_DISP_INFO.ContentRect.top
          dec eax
          invoke MoveToEx, [edi].hdc, dLeft, eax, NULL
          invoke LineTo, [edi].hdc, dLeft, -1
          invoke SelectObject, [edi].hdc, hPenPrev
          call DeleteObject
        .endif

      .else
        ;These are the list box items
        .if [edi].itemData != -1
          mov hFontPrev, $invoke(SelectObject, [edi].hdc, [ebx].hFont)
          mov dPrevBkMode, $invoke(SetBkMode, [edi].hdc, TRANSPARENT)
          .ifBitSet [edi].itemState, ODS_SELECTED
            mov dPrevColor, $invoke(SetTextColor, [edi].hdc, $invoke(GetSysColor, COLOR_HIGHLIGHTTEXT))
            push $invoke(CreateSolidBrush, $invoke(GetSysColor, COLOR_HIGHLIGHT))
          .else
            mov dPrevColor, $invoke(SetTextColor, [edi].hdc, $invoke(GetSysColor, COLOR_WINDOWTEXT))
            push $invoke(CreateSolidBrush, $invoke(GetSysColor, COLOR_WINDOW))
          .endif

          invoke FillRect, [edi].hdc, addr [edi].rcItem, eax
          call DeleteObject

          .if [ebx].dVarType == XPT_VAR_TYPE_SYS_COLOR
            ;Find the matching color index
            push ebx
            push esi
            mov ebx, offset XPT_SYS_COLOR_STR
            mov esi, offset XPT_SYS_COLOR_VAL
            .while byte ptr [ebx] != 0
              .break .if ebx == [edi].itemData          ;Compare pointers
              invoke StrLength, ebx
              add ebx, ecx
              inc esi
            .endw
            movzx eax, byte ptr [esi]                   ;Get the color index
            pop esi
            pop ebx

            ;Render the rect
            invoke GetSysColorBrush, eax                ;Don't delete this brush
            push $invoke(SelectObject, [edi].hdc, eax)
            push [edi].hdc
            mov eax, [edi].rcItem.bottom
            sub eax, XPT_MARGIN
            push eax
            mov eax, [edi].rcItem.left
            add eax, XPT_MARGIN + XPT_COLOR_RECT_SIZE
            push eax
            mov ecx, [edi].rcItem.top
            add ecx, XPT_MARGIN
            push ecx
            sub eax, XPT_COLOR_RECT_SIZE
            push eax
            push [edi].hdc
            call Rectangle
            call SelectObject

            add [edi].rcItem.left, XPT_COLOR_RECT_SIZE + 2*XPT_MARGIN

          .endif
          invoke DrawText, [edi].hdc, [edi].itemData, -1, addr [edi].rcItem, \
                           DT_SINGLELINE or DT_LEFT or DT_END_ELLIPSIS

          ;Restore previous settings
          invoke SelectObject, [edi].hdc, hFontPrev
          invoke SetBkMode, [edi].hdc, dPrevBkMode
          invoke SetTextColor, [edi].hdc, dPrevColor
        .endif
      .endif
      assume ebx:NOTHING, edi:NOTHING

    .elseif [esi].pBtnCtrl != NULL
      mov ecx, [esi].pBtnCtrl
      m2m hWnd, [ecx].WinControl.hWnd
      mov ebx, [ecx].WinControl.pOwner                    ;ebx -> XPT_PropertyNode
      mov pDispInfo, $OCall(esi.GetDispInfo, ebx)
      assume ebx:ptr XPT_PropertyNode, edi:ptr DRAWITEMSTRUCT
      mov ecx, [edi].itemState

      mov hFontPrev, $invoke(SelectObject, [edi].hdc, [ebx].hFont)
      mov dPrevBkMode, $invoke(SetBkMode, [edi].hdc, TRANSPARENT)
      mov dPrevColor, $invoke(SetTextColor, [edi].hdc, $invoke(GetSysColor, COLOR_WINDOWTEXT))
      push $invoke(CreateSolidBrush, $invoke(GetSysColor, COLOR_WINDOW))
      invoke FillRect, [edi].hdc, addr [edi].rcItem, eax
      @invoke DrawText, [edi].hdc, "...", -1, addr [edi].rcItem, \
                       DT_SINGLELINE or DT_CENTER or DT_VCENTER
      call DeleteObject

      ;Restore previous settings
      invoke SelectObject, [edi].hdc, hFontPrev
      invoke SetBkMode, [edi].hdc, dPrevBkMode
      invoke SetTextColor, [edi].hdc, dPrevColor

      assume ebx:NOTHING, edi:NOTHING

    .endif
    mov eax, TRUE
MethodEnd, eax

; 
; Method:    XPropertyTree.OnKeyDown
; Purpose:   Event handler for WM_KEYDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XPropertyTree.OnKeyDown, uses ebx esi, wParam:dword, lParam:dword
    local CRect:RECT

    SetObject esi
    .if [esi].pSelectedNode != NULL
      mov eax, wParam

      .if eax == VK_UP
        ;Move to the above node
        mov ecx, [esi].pSelectedNode
        mov ebx, ecx
        XTV_BrowseBack ecx
        cmp ecx, NULL
        jz @@Exit                                       ;Nothing to do, avoid flicker
        OCall esi.Select, ecx
        OCall esi.GetDispInfo, [esi].pSelectedNode
        .if ecx == 0                                    ;First node in view
          mov [esi].dFirstOffset, 0
          mov eax, [esi].pFirstChild
          .if eax == [esi].pSelectedNode
            mov [esi].TotalOffset.y, 0
          .endif
          OCall esi.Draw
        .elseif eax == NULL                             ;Not in view
          OCall esi.MoveInViewTop, [esi].pSelectedNode
          OCall esi.Draw
        .else
          OCall [esi].pSelectedNode::XTreeViewNode.RenderContent, eax
          OCall esi.GetDispInfo, ebx
          OCall ebx::XTreeViewNode.RenderContent, eax
        .endif

      .elseif eax == VK_DOWN
        ;Move to the below node
        mov ecx, [esi].pSelectedNode
        mov ebx, ecx
        XTV_BrowseFrwd ecx
        cmp ecx, NULL
        jz @@Exit                                       ;Nothing to do, avoid flicker
        OCall esi.Select, ecx
        OCall esi.GetDispInfo, [esi].pSelectedNode
        mov edx, [esi].DispInfoColl.dCount
        dec edx
        .if ecx >= edx                                  ;Is it the last node in view,
          OCall esi.MoveInViewBottom, [esi].pSelectedNode   ;  or is it not in view
          OCall esi.Draw
        .else
          OCall [esi].pSelectedNode::XTreeViewNode.RenderContent, eax
          OCall esi.GetDispInfo, ebx
          OCall ebx::XTreeViewNode.RenderContent, eax
        .endif

      .elseif eax == VK_RIGHT
        ;First expand if collapsed and then move to the first child
        mov ebx, [esi].pSelectedNode
        cmp [ebx].XTreeViewNode.pFirstChild, NULL
        jz @@Exit                                       ;Nothing to do, avoid flicker
        .ifBitClr [ebx].XTreeViewNode.dState, XTVNS_EXPANDED
          OCall ebx::XTreeViewNode.Toggle
        .else
          mov ebx, [ebx].XTreeViewNode.pFirstChild
          OCall esi.Select, ebx
        .endif

        ;Node is now expanded; maybe we have to reposition the tree
        OCall esi.GetDispInfo, ebx
        mov eax, [eax].XTVN_DISP_INFO.sdNodeTop
        add eax, [esi].TotalOffset.y                    ;eax = node offset
        mov edx, eax
        add edx, [ebx].XTreeViewNode.dHeight
        add edx, [ebx].XTreeViewNode.dBranchHeight
        sub edx, [esi].ClientSize.y
        sub edx, [esi].TotalOffset.y
        .if !Sign?
          mov ecx, eax                                  ;The last branch node is off view
          sub ecx, [esi].TotalOffset.y
          sub ecx, edx
          .if !Sign?
            add eax, [ebx].XTreeViewNode.dHeight
            add eax, [ebx].XTreeViewNode.dBranchHeight
            sub eax, [esi].ClientSize.y
          .endif
          OCall esi.Find1stVisibleNode, eax
        .endif
        OCall esi.Draw

      .elseif eax == VK_LEFT
        ;First collapse if expanded and then move to the parent
        mov ecx, [esi].pSelectedNode
        .if XTV_IsExpanded(ecx)
          OCall ecx::XTreeViewNode.Toggle

          ;Correct the client area coords
          mov eax, [esi].TotalSize.x
          .if eax > [esi].ClientSize.x
            invoke ShowScrollBar, [esi].hWnd, SB_HORZ, TRUE
            invoke GetClientRect, [esi].hWnd, addr CRect
            m2m [esi].ClientSize.x, CRect.right
            m2m [esi].ClientSize.y, CRect.bottom
          .endif

          XTV_GetBottomViewNode esi, ecx, eax, edx
          .if ecx == NULL
            OCall esi.MoveInViewBottom, edx
          .endif
        .else
          cmp [ecx].XTreeViewNode.pParent, NULL
          jz @@Exit                                     ;Nothing to do, avoid flicker
          mov ecx, [ecx].XTreeViewNode.pParent
          push ecx
          OCall esi.Select, ecx
          pop ecx
          OCall esi.GetDispInfo, ecx
          .if ecx == 0                                  ;First node in view
            mov [esi].dFirstOffset, 0
          .elseif eax == NULL                           ;Not in view
            OCall esi.MoveInViewTop, [esi].pSelectedNode
          .endif
        .endif
        ;Check if the selected node is in view
        XTV_GetBottomViewNode esi, ecx, eax, edx
        .if ecx == NULL
          OCall esi.MoveInViewBottom, edx
        .endif
        OCall esi.Draw

      .elseif eax == VK_PGUP
        mov eax, [esi].pFirstChild
        cmp eax, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker

        ;Move the selected node on bottom of the view
        OCall esi.MoveInViewBottom, [esi].pSelectedNode

        ;Get the first completely visible node and select it
        mov ecx, [esi].pSelectedNode
        mov edx, [ecx].XTreeViewNode.dHeight
        .repeat
          mov eax, ecx
          XTV_BrowseBack ecx
          .break .if ecx == NULL
          add edx, [ecx].XTreeViewNode.dHeight
        .until edx > [esi].ClientSize.y

        ;Select it and adjust the view if it is the first node
        OCall esi.Select, eax
        XTV_GetBottomViewNode esi, ecx, edx, eax
        .if ecx == NULL
          OCall esi.MoveInViewBottom, eax
        .else
          OCall esi.MoveInViewTop, [esi].pSelectedNode
        .endif
        OCall esi.Draw

      .elseif eax == VK_PGDN
        mov eax, [esi].pLastChild
        cmp eax, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker

        ;Move the selected node on top of the view
        OCall esi.MoveInViewTop, [esi].pSelectedNode

        ;Get the last completely visible node and select it
        mov ecx, [esi].pSelectedNode
        mov edx, [ecx].XTreeViewNode.dHeight
        .repeat
          mov eax, ecx
          XTV_BrowseFrwd ecx
          .break .if ecx == NULL
          add edx, [ecx].XTreeViewNode.dHeight
        .until edx > [esi].ClientSize.y

        ;Select it and adjust the view if it is the last node
        push eax
        push ecx
        OCall esi.Select, eax
        pop ecx
        pop eax
        .if ecx == NULL
          OCall esi.MoveInViewBottom, eax
        .endif
        OCall esi.Draw

      .elseif eax == VK_HOME
        ;Go to the first node in the tree
        mov ecx, [esi].pFirstChild
        cmp ecx, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker
        push ecx
        OCall esi.Set1stVisibleNode, ecx
        pop ecx
        OCall esi.Select, ecx
        mov [esi].TotalOffset.y, 0
        mov [esi].dFirstOffset, 0
        OCall esi.Draw

      .elseif eax == VK_END
        ;Go to the last node in the tree
        mov ecx, [esi].pLastChild
        .while ecx != NULL
          .break .if XTV_IsCollapsed(ecx)
          mov edx, ecx
          mov ecx, [ecx].XTreeViewNode.pLastChild
        .endw
        .if ecx == NULL
          mov ecx, edx
        .endif
        cmp ecx, [esi].pSelectedNode
        je @@Exit                                       ;Nothing to do, avoid flicker
        push ecx
        OCall esi.Select, ecx
        pop ecx
        OCall esi.MoveInViewBottom, ecx
        OCall esi.Draw

      .else
        jmp @@Exit
      .endif
      invoke UpdateWindow, [esi].hWnd
      xor eax, eax

    .else
@@Exit:
      invoke DefWindowProc, [esi].hWnd, WM_KEYDOWN, wParam, lParam

    .endif
MethodEnd, eax

; 
; Method:    XPropertyTree.OnMeasureItem
; Purpose:   Event handler for WM_MEASUREITEM message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XPropertyTree.OnMeasureItem, uses ebx esi, wParam:dword, lParam:dword
    SetObject esi
    ;DbgText "XPropertyTree.OnMeasureItem"
    mov ecx, [esi].pCBxCtrl
    mov ebx, [ecx].WinControl.pOwner
    assume ebx:ptr XPT_PropertyNode
    mov ecx, lParam
    .ifBitSet [ecx].MEASUREITEMSTRUCT.CtlType, ODT_COMBOBOX
      OCall esi.GetDispInfo, ebx
      .if eax != NULL
        mov ecx, lParam
        mov edx, [eax].XPT_DISP_INFO.TextRect.right
        sub edx, [eax].XPT_DISP_INFO.TextRect.left
        mov [ecx].MEASUREITEMSTRUCT.itemWidth, edx
        .if [ecx].MEASUREITEMSTRUCT.itemID == -1    ;Edit control
          mov edx, [eax].XPT_DISP_INFO.sdNodeBot
          sub edx, [eax].XPT_DISP_INFO.sdNodeTop
          sub edx, 6+1
          mov [ecx].MEASUREITEMSTRUCT.itemHeight, edx
        .else
          m2m [ecx].MEASUREITEMSTRUCT.itemHeight, 2*XPT_MARGIN + XPT_COLOR_RECT_SIZE
        .endif
      .endif
    .endif
    assume ebx:NOTHING
    mov eax, TRUE
MethodEnd, eax

; 
; Method:    XPropertyTree.OnMouseWheel
; Purpose:   Event handler for WM_MOUSEWHEEL message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.
; Note:      if the shift key is pressed, the horizontal scroller is moved.

Method XPropertyTree.OnMouseWheel, uses esi, wParam:dword, lParam:dword
    ;DbgText "XPropertyTree.OnMouseWheel"
    SetObject esi
    .ifBitSet [esi].dState, XPTS_CHILD_FOCUSED
      invoke SetFocus, [esi].hWnd
    .endif
    ACall esi.OnMouseWheel, wParam, lParam
MethodEnd, eax

; 
; Method:    XPropertyTree.OnSize
; Purpose:   Event handler for WM_SIZE message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XPropertyTree.OnSize,, wParam:dword, lParam:dword
    SetObject ecx
    mov eax, [ecx].pEdtCtrl                             ;If we have an active control
    or eax, [ecx].pBtnCtrl                              ;  destroy it, changing the focus
    or eax, [ecx].pCBxCtrl                              ;  to the main Wnd
    .if !Zero?
      invoke SetFocus, [ecx].hWnd
    .endif
    ACall OnSize, wParam, lParam
MethodEnd, eax

; 
; Method:    XPropertyTree.OnVScroll
; Purpose:   Event handler for WM_VSCROLL message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XPropertyTree.OnVScroll, uses esi, wParam:dword, lParam:dword
    ;DbgText "XPropertyTree.OnVScroll"
    SetObject esi
    .ifBitSet [esi].dState, XPTS_CHILD_FOCUSED
      invoke SetFocus, [esi].hWnd
    .endif
    ACall esi.OnVScroll, wParam, lParam
MethodEnd

; 
; Method:    XPropertyTree.RenderBackground
; Purpose:   Draws the background of a given rect.
; Arguments: Arg1: -> RECT.
; Return:    Nothing.

Method XPropertyTree.RenderBackground, uses esi, pRect:Pointer
    SetObject esi
    mov eax, pRect
    push [esi].dColorBackgroundDark
    call CreateSolidBrush
    push eax
    push eax
    push pRect
    push [esi].hMemDC
    call FillRect
    call DeleteObject
MethodEnd, NONE

; 
; Method:    XPropertyTree.WindowProc
; Purpose:   Processing of window messages.
; Arguments: Arg1: Message identifier.
;            Arg2: First message parameter.
;            Arg3: Second message parameter.
;            Window Handle is passed in Self.
; Return:    This value is the result of the message processing and depends on the message.

Method XPropertyTree.WndProc, uses esi, uMsg:dword, wParam:dword, lParam:dword
    ;DbgMessage uMsg
    .if uMsg == WM_NCCREATE
      mov eax, lParam
      .if [eax].CREATESTRUCT.lpCreateParams == NULL
        ;if lpCreateParams is 0, then it is a resource call => create new object instance

        mov esi, $New(XPropertyTree, InitRes, Self)
      .else
        ;Init call
        mov esi, [eax].CREATESTRUCT.lpCreateParams
        m2m [esi].XPropertyTree.hWnd, Self
      .endif
      invoke SetWindowLong, Self, GWL_USERDATA, esi
      OCall esi::XPropertyTree.Dispatch, Self, uMsg, wParam, lParam
    .else
      invoke GetWindowLong, Self, GWL_USERDATA              ;if GWL_USERDATA wasn't set,
      .if eax == NULL                                       ;GetWindowLong returns NULL
        invoke DefWindowProc, Self, uMsg, wParam, lParam
      .else
        OCall eax::XPropertyTree.Dispatch, Self, uMsg, wParam, lParam
      .endif
    .endif
MethodEnd, eax

; 
; Method:    XPropertyTree.Startup
; Purpose:   Registers the XPropertyTree object with the OS.
; Arguments: None.
; Return:    Nothing.

Method XPropertyTree.Startup
    local WC:WNDCLASSEX

    mov WC.cbSize, sizeof WNDCLASSEX
    mov WC.style, CS_HREDRAW or CS_VREDRAW or CS_PARENTDC or CS_BYTEALIGNWINDOW or \
                  CS_BYTEALIGNCLIENT or CS_DBLCLKS
    m2m WC.lpfnWndProc, $MethodAddr(XPropertyTree.WndProc)
    mov WC.cbClsExtra, 0
    mov WC.cbWndExtra, 0
    m2m WC.hInstance, hInstance
    mov WC.hbrBackground, 0
    mov WC.lpszMenuName, NULL
    mov WC.lpszClassName, offset szXPT_Class
    mov WC.hIcon, 0
    mov WC.hCursor, $invoke(LoadCursor, hInstance, "CURSOR_ARROW")
    mov WC.hIconSm, 0

    invoke RegisterClassEx, addr WC
MethodEnd, NONE

;##/


; ==================================================================================================
;     XPT_Edit
; ==================================================================================================

;##\

; 
; Method:    XPT_Edit.Done
; Purpose:   Finalizes the XPT_Edit object.
; Arguments: None.
; Return:    Nothing.

Method XPT_Edit.Done, uses esi
    ;DbgText "XPT_Edit.Done"
    SetObject esi
    Unsubclass XPT_Edit                           ;Uses esi
    invoke DestroyWindow, [esi].hWnd
    ACall esi.Done
MethodEnd, NONE

; 
; Method:    XPT_Edit.Init
; Purpose:   Initializes the XPT_Edit object.
; Arguments: Arg1: Pointer to an owner object.
;            Arg2: Handle to a parent window.
;            Arg3: Pointer to an initialization structure.
; Return:    Nothing.

Method XPT_Edit.Init, uses esi, pOwner:Pointer, hParentWnd:Handle, pDefStruc:Pointer
    SetObject esi
    mov ecx, pDefStruc
    assume ecx:ptr XPT_ControlDef
    @invoke CreateWindowEx, [ecx].dExStyle, "EDIT", [ecx].pTitle, [ecx].dStyle, \
            [ecx].sdPosX, [ecx].sdPosY, [ecx].dWidth, [ecx].dHeight, \
            hParentWnd, [ecx].dCtlID, hInstance, 0
    assume ecx:NOTHING
    ACall esi.Init, pOwner, eax
    Subclass XPT_Edit                             ;Uses esi
MethodEnd, NONE

; 
; Method:    XPT_Edit.OnKeyDown
; Purpose:   Event handler for WM_KEYDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XPT_Edit.OnKeyDown, uses edi esi, wParam:dword, lParam:dword
    local bBuffer[20]:byte

    SetObject esi
    mov eax, wParam

    .if eax == VK_ESCAPE
      mov eax, [esi].pOwner
      mov ecx, [eax].XPT_PropertyNode.pOwner
      invoke SetFocus, [ecx].XPropertyTree.hWnd
      xor eax, eax

    .elseif eax == VK_RETURN
      SetObject edi, XPT_PropertyNode, [esi].pOwner
      .if [edi].dVarType == XPT_VAR_TYPE_STRING
        invoke GetWindowText, [esi].hWnd, [edi].pVariable, [edi].pCtrlOptions

      .elseif [edi].dVarType == XPT_VAR_TYPE_INTEGER
        invoke GetWindowText, [esi].hWnd, addr bBuffer, sizeof bBuffer
        invoke StrToSt0, addr bBuffer
        mov eax, [edi].pVariable
        fistp dword ptr [eax]

      .elseif [edi].dVarType == XPT_VAR_TYPE_REAL4
        invoke GetWindowText, [esi].hWnd, addr bBuffer, sizeof bBuffer
        invoke StrToSt0, addr bBuffer
        mov eax, [edi].pVariable
        fstp real4 ptr [eax]

      .endif

      mov ecx, [edi].pOwner
      invoke SetFocus, [ecx].XPropertyTree.hWnd
      xor eax, eax

    .else
      invoke CallWindowProc, [esi].pPrevWndProc, [esi].hWnd, WM_KEYDOWN, wParam, lParam
    .endif
MethodEnd

; 
; Method:    XPT_Edit.WndProc
; Purpose:   Processing of window messages. Before invoking it, the window must be subclassed.
; Arguments: Arg1: Message identifier.
;            Arg2: First message parameter.
;            Arg3: Second message parameter.
;            Window Handle is passed in Self.
; Return:    This value is the result of the message processing and depends on the message.

Method XPT_Edit.WndProc,, uMsg:dword, wParam:dword, lParam:dword
    GetSubclassingInst XPT_Edit, Self
    OCall eax::XPT_Edit.Dispatch, Self, uMsg, wParam, lParam
MethodEnd, eax


;##/


; ==================================================================================================
;     XPT_Button
; ==================================================================================================

;##\

; 
; Method:    XPT_Button.Done
; Purpose:   Finalizes the XPT_Button object.
; Arguments: None.
; Return:    Nothing.

Method XPT_Button.Done, uses esi
    ;DbgText "XPT_Button.Done"
    SetObject esi
    Unsubclass XPT_Button                           ;Uses esi
    invoke DestroyWindow, [esi].hWnd
    ACall esi.Done
MethodEnd, NONE

; 
; Method:    XPT_Button.Init
; Purpose:   Initializes the XPT_Button object.
; Arguments: Arg1: Pointer to an owner object.
;            Arg2: Handle to a parent window.
;            Arg3: Pointer to an initialization structure.
; Return:    Nothing.

Method XPT_Button.Init, uses esi, pOwner:Pointer, hParentWnd:Handle, pDefStruc:Pointer
    SetObject esi
    mov ecx, pDefStruc
    assume ecx:ptr XPT_ControlDef
    @invoke CreateWindowEx, [ecx].dExStyle, "BUTTON", [ecx].pTitle, [ecx].dStyle, \
            [ecx].sdPosX, [ecx].sdPosY, [ecx].dWidth, [ecx].dHeight, \
            hParentWnd, [ecx].dCtlID, hInstance, 0
    assume ecx:NOTHING
    ACall esi.Init, pOwner, eax
    Subclass XPT_Button                             ;Uses esi
MethodEnd, NONE

; 
; Method:    XPT_Button.OnKeyDown
; Purpose:   Event handler for WM_KEYDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XPT_Button.OnKeyDown, uses esi, wParam:dword, lParam:dword
    SetObject esi
    mov eax, wParam
    .if eax == VK_ESCAPE
      mov eax, [esi].pOwner
      mov ecx, [eax].XPT_PropertyNode.pOwner
      invoke SetFocus, [ecx].XPropertyTree.hWnd
      xor eax, eax
    .else
      invoke CallWindowProc, [esi].pPrevWndProc, [esi].hWnd, WM_KEYDOWN, wParam, lParam
    .endif
MethodEnd, eax

; 
; Method:    XPT_Button.WndProc
; Purpose:   Processing of window messages. Before invoking it, the window must be subclassed.
; Arguments: Arg1: Message identifier.
;            Arg2: First message parameter.
;            Arg3: Second message parameter.
;            Window Handle is passed in Self.
; Return:    This value is the result of the message processing and depends on the message.

Method XPT_Button.WndProc,, uMsg:dword, wParam:dword, lParam:dword
    GetSubclassingInst XPT_Button, Self
    OCall eax::XPT_Button.Dispatch, Self, uMsg, wParam, lParam
MethodEnd, eax

;##/


; ==================================================================================================
;     XPT_ComboBox
; ==================================================================================================

;##\

; 
; Method:    XPT_ComboBox.Done
; Purpose:   Finalizes the XPT_ComboBox object.
; Arguments: None.
; Return:    Nothing.

Method XPT_ComboBox.Done, uses esi
    ;DbgText "XPT_ComboBox.Done"
    SetObject esi
    Unsubclass XPT_ComboBox                           ;Uses esi
    invoke DestroyWindow, [esi].hWnd
    ACall esi.Done
MethodEnd, NONE

; 
; Method:    XPT_ComboBox.Init
; Purpose:   Initializes the XPT_ComboBox object.
; Arguments: Arg1: Pointer to an owner object.
;            Arg2: Handle to a parent window.
;            Arg3: Pointer to an initialization structure.
; Return:    Nothing.

Method XPT_ComboBox.Init, uses esi, pOwner:Pointer, hParentWnd:Handle, pDefStruc:Pointer
    SetObject esi
    m2m [esi].pOwner, pOwner                            ;Do this to get the owner in WM_MEASUREITEM
    mov ecx, pDefStruc
    assume ecx:ptr XPT_ControlDef
    @invoke CreateWindowEx, [ecx].dExStyle, "ComboBox", [ecx].pTitle, [ecx].dStyle, \
            [ecx].sdPosX, [ecx].sdPosY, [ecx].dWidth, [ecx].dHeight, \
            hParentWnd, [ecx].dCtlID, hInstance, 0
    assume ecx:NOTHING
    ACall esi.Init, pOwner, eax
    Subclass XPT_ComboBox                               ;Uses esi
MethodEnd, NONE

; 
; Method:    XPT_ComboBox.OnKeyDown
; Purpose:   Event handler for WM_KEYDOWN message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero if handled.

Method XPT_ComboBox.OnKeyDown, uses esi, wParam:dword, lParam:dword
    SetObject esi
    mov eax, wParam
    .if eax == VK_ESCAPE
      mov eax, [esi].pOwner
      mov ecx, [eax].XPT_PropertyNode.pOwner
      BitSet [esi].dState, XPTCB_ESCAPING               ;Set a flag
      invoke SetFocus, [ecx].XPropertyTree.hWnd
      xor eax, eax
    .else
      invoke CallWindowProc, [esi].pPrevWndProc, [esi].hWnd, WM_KEYDOWN, wParam, lParam
    .endif
MethodEnd, eax

; 
; Method:    XPT_ComboBox.OnPaint
; Purpose:   Event handler for WM_PAINT message.
; Arguments: Arg1: First message parameter.
;            Arg2: Second message parameter.
; Return:    Zero.

Method XPT_ComboBox.OnPaint, NOFRAME, wParam:dword, lParam:dword
    xor eax, eax                                        ;Cancel all drawing operations from
MethodEnd 12, eax                                       ;   the original ComboBox

; 
; Method:    XPT_ComboBox.WndProc
; Purpose:   Processing of window messages. Before invoking it, the window must be subclassed.
; Arguments: Arg1: Message identifier.
;            Arg2: First message parameter.
;            Arg3: Second message parameter.
;            Window Handle is passed in Self.
; Return:    This value is the result of the message processing and depends on the message.

Method XPT_ComboBox.WndProc,, uMsg:dword, wParam:dword, lParam:dword
    ;DbgMessage uMsg
    GetSubclassingInst XPT_ComboBox, Self
    OCall eax::XPT_ComboBox.Dispatch, Self, uMsg, wParam, lParam
MethodEnd, eax

;##/

endif

